import { ManagedSocketErrorType } from "./constants";
import { ManagedSocketError } from "./errors";

declare const serializedSocketMessageBrand: unique symbol;

/**
 * Represents a serialized socket message.
 * This type is a branded string, ensuring type safety for serialized socket messages.
 */
export type SerializedSocketMessage = string & {
  [serializedSocketMessageBrand]: true;
};

/**
 * Represents a heartbeat message sent over a socket connection.
 * This message is used to indicate that the connection is still alive.
 */
export interface SocketHeartbeatMessage {
  type: "heartbeat";
  data: string;
}

/**
 * Represents a message with data to be handled by the application
 */
export interface SocketDataMessage {
  type: "message";
  data: unknown;
}

/**
 * Any socket message that can be sent or received.
 */
export type SocketMessage = SocketHeartbeatMessage | SocketDataMessage;

const MESSAGE_TYPES = ["heartbeat", "message"] as const;
export type SocketMessageType = (typeof MESSAGE_TYPES)[number];

/**
 * Parses the given value to determine if it is a valid `SocketMessageType`.
 *
 * @param value - The value to be parsed, expected to be of type `unknown`.
 * @returns The parsed value as `SocketMessageType` if it is valid.
 * @throws {ManagedSocketError} If the value is not a string or not included in `MESSAGE_TYPES`.
 */
export function parseMessageType(value: unknown): SocketMessageType {
  if (
    typeof value !== "string" ||
    !(MESSAGE_TYPES as readonly string[]).includes(value)
  ) {
    throw new ManagedSocketError(
      "Failed to parse incoming message shape",
      new Error("Invalid message type"),
      ManagedSocketErrorType.INVALID_MESSAGE_SHAPE,
    );
  }

  return value as SocketMessageType;
}

/**
 * Parses an incoming socket message and returns a `SocketMessage` object.
 *
 * @param value - The incoming message to parse, expected to be an object.
 * @returns A `SocketMessage` object containing the parsed message type and data.
 * @throws {ManagedSocketError} If the incoming message is not a parseable message
 */
export function parseAnyScoketMessage(value: unknown): SocketMessage {
  if (typeof value !== "object" || value === null) {
    throw new ManagedSocketError(
      "Failed to parse incoming message shape: value was not an object",
      new Error("Invalid message shape"),
      ManagedSocketErrorType.INVALID_MESSAGE_SHAPE,
    );
  }

  if (!("type" in value) || !("data" in value)) {
    throw new ManagedSocketError(
      "Failed to parse incoming message shape: missing type or data",
      new Error("Invalid message shape"),
      ManagedSocketErrorType.INVALID_MESSAGE_SHAPE,
    );
  }

  const { type: incomingType, data: incomingData } = value;
  const type = parseMessageType(incomingType);

  if (type === "heartbeat") {
    if (typeof incomingData !== "string") {
      throw new ManagedSocketError(
        "Failed to parse incoming message shape: heartbeat data was not a string",
        new Error("Invalid message shape"),
        ManagedSocketErrorType.INVALID_MESSAGE_SHAPE,
      );
    }

    return {
      type,
      data: incomingData,
    };
  }

  if (type === "message") {
    return {
      type,
      data: incomingData,
    };
  }

  throw new ManagedSocketError(
    "Failed to parse incoming message shape",
    new Error("Invalid message type"),
    ManagedSocketErrorType.INVALID_MESSAGE_SHAPE,
  );
}
